@@ -47,6 +47,13 @@ and is represented by the special value:  {'__class__':'float','value':'nan'}
 We chose not to encode it as just the string 'NaN' because that could cause
 memory issues when used in multiplications.
 
+Jython's ast module cannot properly parse some literal reprs of unicode strings.
+This is a known bug http://bugs.jython.org/issue2008
+It seems to work when your server is Python 2.x but safest is perhaps to make
+sure your data to parse contains only ascii strings when dealing with Jython.
+Serpent checks for possible problems and will raise an error if it finds one,
+rather than continuing with string data that might be incorrect.
+
 Copyright by Irmen de Jong (irmen@razorvine.net)
 Software license: "MIT software license". See http://opensource.org/licenses/MIT
 """
@@ -59,14 +66,21 @@ import sys
 import types
 import os
 import gc
-import collections
 import decimal
 import datetime
 import uuid
 import array
 import math
+import numbers
+import codecs
+import collections
+if sys.version_info >= (3, 4):
+    from collections.abc import KeysView, ValuesView, ItemsView
+else:
+    from collections import KeysView, ValuesView, ItemsView
 
-__version__ = "1.16"
+
+__version__ = "1.25"
 __all__ = ["dump", "dumps", "load", "loads", "register_class", "unregister_class", "tobytes"]
 
 can_use_set_literals = sys.version_info >= (3, 2)  # check if we can use set literals
@@ -84,17 +98,33 @@ def dump(obj, file, indent=False, set_literals=can_use_set_literals, module_in_c
 
 def loads(serialized_bytes):
     """Deserialize bytes back to object tree. Uses ast.literal_eval (safe)."""
-    serialized = serialized_bytes.decode("utf-8")
-    if sys.version_info < (3, 0) and sys.platform != "cli":
+    if os.name == "java":
+        if type(serialized_bytes) is memoryview:
+            serialized_bytes = serialized_bytes.tobytes()
+        elif type(serialized_bytes) is buffer:
+            serialized_bytes = serialized_bytes[:]
+        serialized = serialized_bytes.decode("utf-8")
+    elif sys.platform == "cli":
+        if type(serialized_bytes) is memoryview:
+            serialized_bytes = serialized_bytes.tobytes()
+        serialized = codecs.decode(serialized_bytes, "utf-8")
+    else:
+        serialized = codecs.decode(serialized_bytes, "utf-8")
+    if '\x00' in serialized:
+        raise ValueError("The serpent data contains 0-bytes so it cannot be parsed by ast.literal_eval. Has it been corrupted?")
+    if sys.version_info < (3, 0):
+        # python 2.x: parse with unicode_literals (promotes all strings to unicode)
+        # note: this doesn't work on jython... see bug http://bugs.jython.org/issue2008
+        # so we add a safety net, to avoid working with incorrectly processed unicode strings
+        serialized = compile(serialized, "<serpent>", mode="eval", flags=ast.PyCF_ONLY_AST | __future__.unicode_literals.compiler_flag)
         if os.name == "java":
-            # Because of a bug in Jython we have to manually convert all Str nodes to unicode. See http://bugs.jython.org/issue2008
-            serialized = ast.parse(serialized, "<serpent>", mode="eval")
             for node in ast.walk(serialized):
-                if isinstance(node, ast.Str) and type(node.s) is str:
-                    node.s = node.s.decode("utf-8")
-        else:
-            # python 2.x: parse with unicode_literals (promotes all strings to unicode)
-            serialized = compile(serialized, "<serpent>", mode="eval", flags=ast.PyCF_ONLY_AST | __future__.unicode_literals.compiler_flag)
+                if isinstance(node, ast.Str):
+                    if isinstance(node.s, str) and any(c for c in node.s if c > '\x7f'):
+                        # In this case there is risk of incorrectly parsed unicode data. Play safe and crash.
+                        raise ValueError("cannot properly parse unicode string with ast in Jython, see bug http://bugs.jython.org/issue2008"
+                                         " - use python 2.x server or convert strings to ascii yourself first")
+                    node.s = node.s.decode("unicode-escape")
     try:
         if os.name != "java" and sys.platform != "cli":
             gc.disable()
@@ -121,18 +151,25 @@ def _ser_DictView(obj, serializer, outputstream, indentlevel):
     serializer.ser_builtins_list(obj, outputstream, indentlevel)
 
 
-_special_classes_registry = {
-    collections.KeysView: _ser_DictView,
-    collections.ValuesView: _ser_DictView,
-    collections.ItemsView: _ser_DictView
-}
-if sys.version_info >= (2, 7):
-    _special_classes_registry[collections.OrderedDict] = _ser_OrderedDict
-if sys.version_info >= (3, 4):
-    import enum
-    def _ser_Enum(obj, serializer, outputstream, indentlevel):
-        serializer._serialize(obj.value, outputstream, indentlevel)
-    _special_classes_registry[enum.Enum] = _ser_Enum
+_special_classes_registry = collections.OrderedDict()   # must be insert-order preserving to make sure of proper precedence rules
+
+
+def _reset_special_classes_registry():
+    _special_classes_registry.clear()
+    _special_classes_registry[KeysView] = _ser_DictView
+    _special_classes_registry[ValuesView] = _ser_DictView
+    _special_classes_registry[ItemsView] = _ser_DictView
+    if sys.version_info >= (2, 7):
+        _special_classes_registry[collections.OrderedDict] = _ser_OrderedDict
+    if sys.version_info >= (3, 4):
+        import enum
+
+        def _ser_Enum(obj, serializer, outputstream, indentlevel):
+            serializer._serialize(obj.value, outputstream, indentlevel)
+        _special_classes_registry[enum.Enum] = _ser_Enum
+
+
+_reset_special_classes_registry()
 
 
 def unregister_class(clazz):
@@ -187,13 +224,6 @@ class BytesWrapper(object):
         return BytesWrapper(data)
 
 
-if sys.version_info < (3, 0):
-    _repr = repr     # python <3.0 won't need explicit encoding to utf-8, so we optimize this
-else:
-    def _repr(obj):
-        return repr(obj).encode("utf-8")
-
-
 _repr_types = set([
     str,
     int,
@@ -241,7 +271,10 @@ def tobytes(obj):
     if isinstance(obj, _bytes_types):
         return obj
     if isinstance(obj, dict) and "data" in obj and obj.get("encoding") == "base64":
-        return base64.b64decode(obj["data"])
+        try:
+            return base64.b64decode(obj["data"])
+        except TypeError:
+            return base64.b64decode(obj["data"].encode("ascii"))   # needed for certain older versions of pypy
     raise TypeError("argument is neither bytes nor serpent base64 encoded bytes dict")
 
 
@@ -258,7 +291,8 @@ class Serializer(object):
         """
         Initialize the serializer.
         indent=indent the output over multiple lines (default=false)
-        setLiterals=use set-literals or not (set to False if you need compatibility with Python < 3.2). Serpent chooses a sensible default for you.
+        setLiterals=use set-literals or not (set to False if you need compatibility with Python < 3.2).
+        Serpent chooses a sensible default for you.
         module_in_classname = include module prefix for class names or only use the class name itself
         """
         self.indent = indent
@@ -266,6 +300,7 @@ class Serializer(object):
         self.module_in_classname = module_in_classname
         self.serialized_obj_ids = set()
         self.special_classes_registry_copy = None
+        self.maximum_level = min(sys.getrecursionlimit() // 5, 1000)
 
     def serialize(self, obj):
         """Serialize the object tree to bytes."""
@@ -274,8 +309,10 @@ class Serializer(object):
         if self.set_literals:
             header += "python3.2\n"   # set-literals require python 3.2+ to deserialize (ast.literal_eval limitation)
         else:
-            header += "python2.6\n"   # don't change this even though we don't support 2.6 any longer, otherwise we can't read older serpent strings
-        out = [header.encode("utf-8")]
+            header += "python2.6\n"   # don't change this, otherwise we can't read older serpent strings
+        out = [header]
+        if os.name == "java" and type(obj) is buffer:
+            obj = bytearray(obj)
         try:
             if os.name != "java" and sys.platform != "cli":
                 gc.disable()
@@ -285,19 +322,20 @@ class Serializer(object):
             gc.enable()
         self.special_classes_registry_copy = None
         del self.serialized_obj_ids
-        if sys.platform == "cli":
-            return "".join(out)
-        return b"".join(out)
+        return "".join(out).encode("utf-8")
 
     _shortcut_dispatch_types = frozenset([float, complex, tuple, list, dict, set, frozenset])
 
     def _serialize(self, obj, out, level):
+        if level > self.maximum_level:
+            raise ValueError("Object graph nesting too deep. Increase serializer.maximum_level if you think you need more, "
+                             " but this may cause a RecursionError instead if Python's recursion limit doesn't allow it.")
         t = type(obj)
         if t in _translate_types:
             obj = _translate_types[t](obj)
             t = type(obj)
         if t in _repr_types:
-            out.append(_repr(obj))    # just a simple repr() is enough for these objects
+            out.append(repr(obj))    # just a simple repr() is enough for these objects
             return
         if t in self._shortcut_dispatch_types:
             # we shortcut these builtins directly to the dispatch function to avoid type lookup overhead below
@@ -330,47 +368,32 @@ class Serializer(object):
     def ser_builtins_float(self, float_obj, out, level):
         if math.isnan(float_obj):
             # there's no literal expression for a float NaN...
-            out.append(b"{'__class__':'float','value':'nan'}")
+            out.append("{'__class__':'float','value':'nan'}")
         elif math.isinf(float_obj):
             # output a literal expression that overflows the float and results in +/-INF
             if float_obj > 0:
-                out.append(b"1e30000")
+                out.append("1e30000")
             else:
-                out.append(b"-1e30000")
+                out.append("-1e30000")
         else:
-            out.append(repr(float_obj).encode("ascii"))
+            out.append(repr(float_obj))
     dispatch[float] = ser_builtins_float
 
     def ser_builtins_complex(self, complex_obj, out, level):
-        out.append(b"(")
+        out.append("(")
         self.ser_builtins_float(complex_obj.real, out, level)
         if complex_obj.imag >= 0:
-            out.append(b"+")
+            out.append("+")
         self.ser_builtins_float(complex_obj.imag, out, level)
-        out.append(b"j)")
+        out.append("j)")
     dispatch[complex] = ser_builtins_complex
 
     if sys.version_info < (3, 0):
+        # this method is used for python 2.x unicode (python 3.x doesn't use this)
         def ser_builtins_unicode(self, unicode_obj, out, level):
-            # this method is used for python 2.x unicode (python 3.x doesn't use this)
-            z = unicode_obj.encode("utf-8")
-            # double-escape existing backslashes:
-            z = z.replace("\\", "\\\\")
-            # backslash-escape control characters:
-            z = z.replace("\a", "\\a")
-            z = z.replace("\b", "\\b")
-            z = z.replace("\f", "\\f")
-            z = z.replace("\n", "\\n")
-            z = z.replace("\r", "\\r")
-            z = z.replace("\t", "\\t")
-            z = z.replace("\v", "\\v")
-            if "'" not in z:
-                z = "'" + z + "'"
-            elif '"' not in z:
-                z = '"' + z + '"'
-            else:
-                z = z.replace("'", "\\'")
-                z = "'" + z + "'"
+            z = repr(unicode_obj)
+            if z[0] == 'u':
+                z = z[1:]    # get rid of the unicode 'u' prefix
             out.append(z)
         dispatch[unicode] = ser_builtins_unicode
 
@@ -384,25 +407,25 @@ class Serializer(object):
         append = out.append
         serialize = self._serialize
         if self.indent and tuple_obj:
-            indent_chars = b"  " * level
-            indent_chars_inside = indent_chars + b"  "
-            append(b"(\n")
+            indent_chars = "  " * level
+            indent_chars_inside = indent_chars + "  "
+            append("(\n")
             for elt in tuple_obj:
                 append(indent_chars_inside)
                 serialize(elt, out, level + 1)
-                append(b",\n")
+                append(",\n")
             out[-1] = out[-1].rstrip()  # remove the last \n
             if len(tuple_obj) > 1:
                 del out[-1]  # undo the last ,
-            append(b"\n" + indent_chars + b")")
+            append("\n" + indent_chars + ")")
         else:
-            append(b"(")
+            append("(")
             for elt in tuple_obj:
                 serialize(elt, out, level + 1)
-                append(b",")
+                append(",")
             if len(tuple_obj) > 1:
                 del out[-1]  # undo the last ,
-            append(b")")
+            append(")")
     dispatch[tuple] = ser_builtins_tuple
 
     def ser_builtins_list(self, list_obj, out, level):
@@ -412,26 +435,37 @@ class Serializer(object):
         append = out.append
         serialize = self._serialize
         if self.indent and list_obj:
-            indent_chars = b"  " * level
-            indent_chars_inside = indent_chars + b"  "
-            append(b"[\n")
+            indent_chars = "  " * level
+            indent_chars_inside = indent_chars + "  "
+            append("[\n")
             for elt in list_obj:
                 append(indent_chars_inside)
                 serialize(elt, out, level + 1)
-                append(b",\n")
+                append(",\n")
             del out[-1]  # remove the last ,\n
-            append(b"\n" + indent_chars + b"]")
+            append("\n" + indent_chars + "]")
         else:
-            append(b"[")
+            append("[")
             for elt in list_obj:
                 serialize(elt, out, level + 1)
-                append(b",")
+                append(",")
             if list_obj:
                 del out[-1]  # remove the last ,
-            append(b"]")
+            append("]")
         self.serialized_obj_ids.discard(id(list_obj))
     dispatch[list] = ser_builtins_list
 
+    def _check_hashable_type(self, t):
+        if t not in (bool, bytes, str, tuple) and not issubclass(t, numbers.Number):
+            if sys.version_info >= (3, 4):
+                import enum
+                if issubclass(t, enum.Enum):
+                    return
+            elif sys.version_info < (3, 0) and t is unicode:
+                return
+            raise TypeError("one of the keys in a dict or set is not of a primitive hashable type: " +
+                            str(t) + ". Use simple types as keys or use a list or tuple as container.")
+
     def ser_builtins_dict(self, dict_obj, out, level):
         if id(dict_obj) in self.serialized_obj_ids:
             raise ValueError("Circular reference detected (dict)")
@@ -439,9 +473,9 @@ class Serializer(object):
         append = out.append
         serialize = self._serialize
         if self.indent and dict_obj:
-            indent_chars = b"  " * level
-            indent_chars_inside = indent_chars + b"  "
-            append(b"{\n")
+            indent_chars = "  " * level
+            indent_chars_inside = indent_chars + "  "
+            append("{\n")
             dict_items = dict_obj.items()
             try:
                 sorted_items = sorted(dict_items)
@@ -449,22 +483,24 @@ class Serializer(object):
                 sorted_items = dict_items
             for key, value in sorted_items:
                 append(indent_chars_inside)
+                self._check_hashable_type(type(key))
                 serialize(key, out, level + 1)
-                append(b": ")
+                append(": ")
                 serialize(value, out, level + 1)
-                append(b",\n")
+                append(",\n")
             del out[-1]  # remove last ,\n
-            append(b"\n" + indent_chars + b"}")
+            append("\n" + indent_chars + "}")
         else:
-            append(b"{")
+            append("{")
             for key, value in dict_obj.items():
+                self._check_hashable_type(type(key))
                 serialize(key, out, level + 1)
-                append(b":")
+                append(":")
                 serialize(value, out, level + 1)
-                append(b",")
+                append(",")
             if dict_obj:
                 del out[-1]  # remove the last ,
-            append(b"}")
+            append("}")
         self.serialized_obj_ids.discard(id(dict_obj))
     dispatch[dict] = ser_builtins_dict
 
@@ -477,26 +513,28 @@ class Serializer(object):
         append = out.append
         serialize = self._serialize
         if self.indent and set_obj:
-            indent_chars = b"  " * level
-            indent_chars_inside = indent_chars + b"  "
-            append(b"{\n")
+            indent_chars = "  " * level
+            indent_chars_inside = indent_chars + "  "
+            append("{\n")
             try:
                 sorted_elts = sorted(set_obj)
             except TypeError:   # can occur when elements can't be ordered (Python 3.x)
                 sorted_elts = set_obj
             for elt in sorted_elts:
                 append(indent_chars_inside)
+                self._check_hashable_type(type(elt))
                 serialize(elt, out, level + 1)
-                append(b",\n")
+                append(",\n")
             del out[-1]  # remove the last ,\n
-            append(b"\n" + indent_chars + b"}")
+            append("\n" + indent_chars + "}")
         elif set_obj:
-            append(b"{")
+            append("{")
             for elt in set_obj:
+                self._check_hashable_type(type(elt))
                 serialize(elt, out, level + 1)
-                append(b",")
+                append(",")
             del out[-1]  # remove the last ,
-            append(b"}")
+            append("}")
         else:
             # empty set literal doesn't exist unfortunately, replace with empty tuple
             self.ser_builtins_tuple((), out, level)
@@ -508,33 +546,33 @@ class Serializer(object):
 
     def ser_decimal_Decimal(self, decimal_obj, out, level):
         # decimal is serialized as a string to avoid losing precision
-        self._serialize(str(decimal_obj), out, level)
+        out.append(repr(str(decimal_obj)))
     dispatch[decimal.Decimal] = ser_decimal_Decimal
 
     def ser_datetime_datetime(self, datetime_obj, out, level):
-        self._serialize(datetime_obj.isoformat(), out, level)
+        out.append(repr(datetime_obj.isoformat()))
     dispatch[datetime.datetime] = ser_datetime_datetime
-    
+
     def ser_datetime_date(self, date_obj, out, level):
-        self._serialize(date_obj.isoformat(), out, level)
+        out.append(repr(date_obj.isoformat()))
     dispatch[datetime.date] = ser_datetime_date
-    
+
     if os.name == "java" or sys.version_info < (2, 7):    # jython bug http://bugs.jython.org/issue2010
         def ser_datetime_timedelta(self, timedelta_obj, out, level):
             secs = ((timedelta_obj.days * 86400 + timedelta_obj.seconds) * 10 ** 6 + timedelta_obj.microseconds) / 10 ** 6
-            self._serialize(secs, out, level)
+            out.append(repr(secs))
     else:
         def ser_datetime_timedelta(self, timedelta_obj, out, level):
             secs = timedelta_obj.total_seconds()
-            self._serialize(secs, out, level)
+            out.append(repr(secs))
     dispatch[datetime.timedelta] = ser_datetime_timedelta
 
     def ser_datetime_time(self, time_obj, out, level):
-        self._serialize(str(time_obj), out, level)
+        out.append(repr(str(time_obj)))
     dispatch[datetime.time] = ser_datetime_time
 
     def ser_uuid_UUID(self, uuid_obj, out, level):
-        self._serialize(str(uuid_obj), out, level)
+        out.append(repr(str(uuid_obj)))
     dispatch[uuid.UUID] = ser_uuid_UUID
 
     def ser_exception_class(self, exc_obj, out, level):
@@ -584,7 +622,8 @@ class Serializer(object):
                             value[slot] = getattr(obj, slot)
                         value["__class__"] = self.get_class_name(obj)
                     else:
-                        raise TypeError("don't know how to serialize class " + str(obj.__class__) + ". Give it vars() or an appropriate __getstate__")
+                        raise TypeError("don't know how to serialize class " +
+                                        str(obj.__class__) + ". Give it vars() or an appropriate __getstate__")
             self._serialize(value, out, level)
         finally:
             self.serialized_obj_ids.discard(id(obj))
